/*
 *  Copyright 2001-2013 Stephen Colebourne
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.joda.beans.gen;

import java.util.Map;

import org.joda.beans.Bean;
import org.joda.beans.BeanBuilder;
import org.joda.beans.BeanDefinition;
import org.joda.beans.DerivedProperty;
import org.joda.beans.JodaBeanUtils;
import org.joda.beans.MetaProperty;
import org.joda.beans.Property;
import org.joda.beans.PropertyDefinition;
import org.joda.beans.impl.direct.DirectBean;
import org.joda.beans.impl.direct.DirectBeanBuilder;
import org.joda.beans.impl.direct.DirectMetaBean;
import org.joda.beans.impl.direct.DirectMetaProperty;
import org.joda.beans.impl.direct.DirectMetaPropertyMap;

/**
 * Mock used for test equals.
 * 
 * @author Stephen Colebourne
 */
@BeanDefinition
public class RWOnlyBean extends DirectBean {

    /**
     * The read only property.
     */
    @PropertyDefinition(set = "")
    private Object ro = null;  // trailing comment
    /**
     * The write only property.
     */
    @PropertyDefinition(get = "")
    private Object wo;  // trailing comment
    /**
     * The final read only property.
     */
    @PropertyDefinition
    private final Object fin;
    /**
     * The private get/set property.
     */
    @PropertyDefinition(get = "private", set = "private")
    private String priv;

    public RWOnlyBean() {
        fin = "";
    }

    /**
     * A manual get property, no set.
     */
    @PropertyDefinition(get = "manual", set = "")
    private String manualGet;

    public String getManualGet() {
        return "goo";
    }

    /**
     * Gets the value of a derived property.
     * 
     * @return derived
     */
    @DerivedProperty
    public String getDerived() {
        return "drv";
    }

    //------------------------- AUTOGENERATED START -------------------------
    ///CLOVER:OFF
    /**
     * The meta-bean for {@code RWOnlyBean}.
     * @return the meta-bean, not null
     */
    public static RWOnlyBean.Meta meta() {
        return RWOnlyBean.Meta.INSTANCE;
    }

    static {
        JodaBeanUtils.registerMetaBean(RWOnlyBean.Meta.INSTANCE);
    }

    @Override
    public RWOnlyBean.Meta metaBean() {
        return RWOnlyBean.Meta.INSTANCE;
    }

    //-----------------------------------------------------------------------
    /**
     * Gets the read only property.
     * @return the value of the property
     */
    public Object getRo() {
        return ro;
    }

    /**
     * Gets the the {@code ro} property.
     * @return the property, not null
     */
    public final Property<Object> ro() {
        return metaBean().ro().createProperty(this);
    }

    //-----------------------------------------------------------------------
    /**
     * Sets the write only property.
     * @param wo  the new value of the property
     */
    public void setWo(Object wo) {
        this.wo = wo;
    }

    /**
     * Gets the the {@code wo} property.
     * @return the property, not null
     */
    public final Property<Object> wo() {
        return metaBean().wo().createProperty(this);
    }

    //-----------------------------------------------------------------------
    /**
     * Gets the final read only property.
     * @return the value of the property
     */
    public Object getFin() {
        return fin;
    }

    /**
     * Gets the the {@code fin} property.
     * @return the property, not null
     */
    public final Property<Object> fin() {
        return metaBean().fin().createProperty(this);
    }

    //-----------------------------------------------------------------------
    /**
     * Gets the private get/set property.
     * @return the value of the property
     */
    private String getPriv() {
        return priv;
    }

    /**
     * Sets the private get/set property.
     * @param priv  the new value of the property
     */
    private void setPriv(String priv) {
        this.priv = priv;
    }

    /**
     * Gets the the {@code priv} property.
     * @return the property, not null
     */
    public final Property<String> priv() {
        return metaBean().priv().createProperty(this);
    }

    //-----------------------------------------------------------------------
    /**
     * Gets the the {@code manualGet} property.
     * @return the property, not null
     */
    public final Property<String> manualGet() {
        return metaBean().manualGet().createProperty(this);
    }

    //-----------------------------------------------------------------------
    /**
     * Gets the the {@code derived} property.
     * 
     * @return the property, not null
     */
    public final Property<String> derived() {
        return metaBean().derived().createProperty(this);
    }

    //-----------------------------------------------------------------------
    @Override
    public RWOnlyBean clone() {
        BeanBuilder<? extends RWOnlyBean> builder = metaBean().builder();
        for (MetaProperty<?> mp : metaBean().metaPropertyIterable()) {
            if (mp.style().isBuildable()) {
                Object value = mp.get(this);
                if (value instanceof Bean) {
                    value = ((Bean) value).clone();
                }
                builder.set(mp.name(), value);
            }
        }
        return builder.build();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }
        if (obj != null && obj.getClass() == this.getClass()) {
            RWOnlyBean other = (RWOnlyBean) obj;
            return JodaBeanUtils.equal(getRo(), other.getRo()) &&
                    JodaBeanUtils.equal(wo, other.wo) &&
                    JodaBeanUtils.equal(getFin(), other.getFin()) &&
                    JodaBeanUtils.equal(getPriv(), other.getPriv()) &&
                    JodaBeanUtils.equal(getManualGet(), other.getManualGet()) &&
                    JodaBeanUtils.equal(getDerived(), other.getDerived());
        }
        return false;
    }

    @Override
    public int hashCode() {
        int hash = getClass().hashCode();
        hash += hash * 31 + JodaBeanUtils.hashCode(getRo());
        hash += hash * 31 + JodaBeanUtils.hashCode(wo);
        hash += hash * 31 + JodaBeanUtils.hashCode(getFin());
        hash += hash * 31 + JodaBeanUtils.hashCode(getPriv());
        hash += hash * 31 + JodaBeanUtils.hashCode(getManualGet());
        hash += hash * 31 + JodaBeanUtils.hashCode(getDerived());
        return hash;
    }

    @Override
    public String toString() {
        StringBuilder buf = new StringBuilder(224);
        buf.append("RWOnlyBean{");
        int len = buf.length();
        toString(buf);
        if (buf.length() > len) {
            buf.setLength(buf.length() - 2);
        }
        buf.append('}');
        return buf.toString();
    }

    protected void toString(StringBuilder buf) {
        buf.append("ro").append('=').append(JodaBeanUtils.toString(getRo())).append(',').append(' ');
        buf.append("wo").append('=').append(JodaBeanUtils.toString(wo)).append(',').append(' ');
        buf.append("fin").append('=').append(JodaBeanUtils.toString(getFin())).append(',').append(' ');
        buf.append("priv").append('=').append(JodaBeanUtils.toString(getPriv())).append(',').append(' ');
        buf.append("manualGet").append('=').append(JodaBeanUtils.toString(getManualGet())).append(',').append(' ');
        buf.append("derived").append('=').append(JodaBeanUtils.toString(getDerived())).append(',').append(' ');
    }

    //-----------------------------------------------------------------------
    /**
     * The meta-bean for {@code RWOnlyBean}.
     */
    public static class Meta extends DirectMetaBean {
        /**
         * The singleton instance of the meta-bean.
         */
        static final Meta INSTANCE = new Meta();

        /**
         * The meta-property for the {@code ro} property.
         */
        private final MetaProperty<Object> ro = DirectMetaProperty.ofReadOnly(
                this, "ro", RWOnlyBean.class, Object.class);
        /**
         * The meta-property for the {@code wo} property.
         */
        private final MetaProperty<Object> wo = DirectMetaProperty.ofWriteOnly(
                this, "wo", RWOnlyBean.class, Object.class);
        /**
         * The meta-property for the {@code fin} property.
         */
        private final MetaProperty<Object> fin = DirectMetaProperty.ofReadOnly(
                this, "fin", RWOnlyBean.class, Object.class);
        /**
         * The meta-property for the {@code priv} property.
         */
        private final MetaProperty<String> priv = DirectMetaProperty.ofReadWrite(
                this, "priv", RWOnlyBean.class, String.class);
        /**
         * The meta-property for the {@code manualGet} property.
         */
        private final MetaProperty<String> manualGet = DirectMetaProperty.ofReadOnly(
                this, "manualGet", RWOnlyBean.class, String.class);
        /**
         * The meta-property for the {@code derived} property.
         */
        private final MetaProperty<String> derived = DirectMetaProperty.ofDerived(
                this, "derived", RWOnlyBean.class, String.class);
        /**
         * The meta-properties.
         */
        private final Map<String, MetaProperty<?>> metaPropertyMap$ = new DirectMetaPropertyMap(
                this, null,
                "ro",
                "wo",
                "fin",
                "priv",
                "manualGet",
                "derived");

        /**
         * Restricted constructor.
         */
        protected Meta() {
        }

        @Override
        protected MetaProperty<?> metaPropertyGet(String propertyName) {
            switch (propertyName.hashCode()) {
                case 3645:  // ro
                    return ro;
                case 3800:  // wo
                    return wo;
                case 101387:  // fin
                    return fin;
                case 3449519:  // priv
                    return priv;
                case 93508016:  // manualGet
                    return manualGet;
                case 1556125213:  // derived
                    return derived;
            }
            return super.metaPropertyGet(propertyName);
        }

        @Override
        public BeanBuilder<? extends RWOnlyBean> builder() {
            return new DirectBeanBuilder<RWOnlyBean>(new RWOnlyBean());
        }

        @Override
        public Class<? extends RWOnlyBean> beanType() {
            return RWOnlyBean.class;
        }

        @Override
        public Map<String, MetaProperty<?>> metaPropertyMap() {
            return metaPropertyMap$;
        }

        //-----------------------------------------------------------------------
        /**
         * The meta-property for the {@code ro} property.
         * @return the meta-property, not null
         */
        public final MetaProperty<Object> ro() {
            return ro;
        }

        /**
         * The meta-property for the {@code wo} property.
         * @return the meta-property, not null
         */
        public final MetaProperty<Object> wo() {
            return wo;
        }

        /**
         * The meta-property for the {@code fin} property.
         * @return the meta-property, not null
         */
        public final MetaProperty<Object> fin() {
            return fin;
        }

        /**
         * The meta-property for the {@code priv} property.
         * @return the meta-property, not null
         */
        public final MetaProperty<String> priv() {
            return priv;
        }

        /**
         * The meta-property for the {@code manualGet} property.
         * @return the meta-property, not null
         */
        public final MetaProperty<String> manualGet() {
            return manualGet;
        }

        /**
         * The meta-property for the {@code derived} property.
         * @return the meta-property, not null
         */
        public final MetaProperty<String> derived() {
            return derived;
        }

        //-----------------------------------------------------------------------
        @Override
        protected Object propertyGet(Bean bean, String propertyName, boolean quiet) {
            switch (propertyName.hashCode()) {
                case 3645:  // ro
                    return ((RWOnlyBean) bean).getRo();
                case 3800:  // wo
                    if (quiet) {
                        return null;
                    }
                    throw new UnsupportedOperationException("Property cannot be read: wo");
                case 101387:  // fin
                    return ((RWOnlyBean) bean).getFin();
                case 3449519:  // priv
                    return ((RWOnlyBean) bean).getPriv();
                case 93508016:  // manualGet
                    return ((RWOnlyBean) bean).getManualGet();
                case 1556125213:  // derived
                    return ((RWOnlyBean) bean).getDerived();
            }
            return super.propertyGet(bean, propertyName, quiet);
        }

        @Override
        protected void propertySet(Bean bean, String propertyName, Object newValue, boolean quiet) {
            switch (propertyName.hashCode()) {
                case 3645:  // ro
                    if (quiet) {
                        return;
                    }
                    throw new UnsupportedOperationException("Property cannot be written: ro");
                case 3800:  // wo
                    ((RWOnlyBean) bean).setWo((Object) newValue);
                    return;
                case 101387:  // fin
                    if (quiet) {
                        return;
                    }
                    throw new UnsupportedOperationException("Property cannot be written: fin");
                case 3449519:  // priv
                    ((RWOnlyBean) bean).setPriv((String) newValue);
                    return;
                case 93508016:  // manualGet
                    if (quiet) {
                        return;
                    }
                    throw new UnsupportedOperationException("Property cannot be written: manualGet");
                case 1556125213:  // derived
                    if (quiet) {
                        return;
                    }
                    throw new UnsupportedOperationException("Property cannot be written: derived");
            }
            super.propertySet(bean, propertyName, newValue, quiet);
        }

    }

    ///CLOVER:ON
    //-------------------------- AUTOGENERATED END --------------------------
}
